"""
Mapper class for mapping a graph into a subgraph of another.

Author: Per Erik Strandberg, 2017

The source code in this bundle explores the subgraph isomorphism problem
Copyright (C) 2018 P. E. Strandberg, Westermo Research and Development AB

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""

from network import Graph, GraphException
from copy import deepcopy
from tree import Tree
from datetime import datetime


class Mapper:
    """Class for mapping a potential subgraph H onto a graph G."""

    def __init__(self, G, H, find_all=True, check_old=False,
                 old_ones=None, timeout=None):
        """Constructor just stores G and H."""
        self.G = deepcopy(G)
        self.H = deepcopy(H)
        self.candidates = dict()
        self.not_candidates = self.G.get_nodes()
        self.solutions = list()
        self.check_old = None

        self.searchtree = None

        self.ctr = 0
        self.ctrmax = 1000000
        self.started = datetime.now()
        self.maxtime = 45  # seconds
        if timeout and timeout > 0:
            print "max time:",
            print timeout

            self.maxtime = timeout

        # 1 -- preprocessing of G - remove nodes that no-one wants to map to
        self.get_all_candidates()

        # 2 -- unmappability tests
        self.check_simple_unmappability()
        self.check_loop_unmappability()

        # 3 -- order nodes according to search strategy
        if check_old:
            self.old_ones = old_ones
            self.setup_search_strategy3()
        else:
            self.setup_search_strategy()

        # 4 -- build the mapping
        self.solutions = self.search(find_all)

        return

    def get_candidates(self, hnode):
        """Get the candidates of one node."""
        hdeg = self.H.get_degree(hnode)

        self.candidates[hnode] = list()

        for gnode in self.G.get_nodes():
            if self.G.get_degree(gnode) >= hdeg:
                self.candidates[hnode].append(gnode)
                if gnode in self.not_candidates:
                    self.not_candidates.remove(gnode)

        print "* current candidates of %s are %s" % (hnode,
                                                     self.candidates[hnode])

        return

    def get_all_candidates(self):
        """Setup the candidates-nodes in G, of the nodes in H."""
        print "*** get all candidates ***"

        for hnode in self.H.get_nodes():
            self.get_candidates(hnode)

        # this is the naive way of doing it
        # TODO: implement something that does not redo this all the time
        if self.not_candidates:

            print "some nodes of G can be removed:",
            print self.not_candidates

            print "%s nodes left" % (len(self.G.get_nodes()) -
                                     len(self.not_candidates))

            while self.not_candidates:
                gnode = self.not_candidates.pop()
                self.G.del_node(gnode)

            self.not_candidates = self.G.get_nodes()
            self.get_all_candidates()

        return False

    def check_simple_unmappability(self):
        """Check if H cannot be mappable to G. Return True if unmappable."""
        # number of nodes
        if self.G.get_num_nodes() < self.H.get_num_nodes():
            msg = "Not enough nodes in G (%d vs %d)."
            msg = msg % (self.G.get_num_nodes(), self.H.get_num_nodes())
            print msg
            raise GraphException(msg)

        # number of edges
        if self.G.get_num_edges() < self.H.get_num_edges():
            msg = "Not enough edges in G (%d vs %d)."
            msg = msg % (self.G.get_num_edges(), self.H.get_num_edges())
            print msg
            raise GraphException(msg)

        # check degree sequence -- the i'th most connected node in G
        # cannot be less connected than the i'th node in H.
        hseq = self.H.get_deg_seq()
        gseq = self.G.get_deg_seq()

        for i in xrange(self.H.get_num_nodes()):
            if gseq[i] < hseq[i]:
                msg = "Degree sequence check, failed. "
                msg += "G: %s. " % gseq
                msg += "H: %s." % hseq
                print msg
                raise GraphException(msg)

        # empty candidates
        for hnode, cands in self.candidates.iteritems():
            if len(cands) == 0:
                msg = "Node %s has no candidates." % hnode
                print msg
                raise GraphException(msg)

        # pigeon hole principle
        # TODO: Implement me

        return

    def check_loop_unmappability(self):
        """Check loops of odd length in H and G."""
        # This might sound hard to believe, but we need to treat
        # the graph like a matrix for a while. The wikipedia article
        # on this is a good starting point, especially the section on
        # Matrix Powers. https://en.wikipedia.org/wiki/Adjacency_matrix

        A = self.G.get_matrix()
        A2 = A * A

        B = self.H.get_matrix()
        B2 = B * B

        Ak = A
        Bk = B
        k = 1

        while (k < len(self.H.get_nodes())) and (k < 22):
            k += 2

            Ak = Ak * A2
            Bk = Bk * B2

            # count number of nodes involved in loop
            gnodes = list()
            hnodes = list()

            for idx in xrange(len(self.G.get_nodes())):
                if Ak[idx, idx] > 0:
                    gnodes.append(self.G.get_nodes()[idx])

            for idx in xrange(len(self.H.get_nodes())):
                if Bk[idx, idx] > 0:
                    hnodes.append(self.H.get_nodes()[idx])

            if len(hnodes) > len(gnodes):
                msg = "There are more loops of length %s in H than G :-(" % k
                print msg
                raise GraphException(msg)

            # A node not in a loop of length k in G cannot be in a
            # loop of length k in H, so in that case we remove it.
            # (In a simple test case this gave a speed-up of 1000
            # times.)
            for hnode in hnodes:
                for candidate in deepcopy(self.candidates[hnode]):
                    if candidate not in gnodes:
                        self.candidates[hnode].remove(candidate)

        return

    def setup_search_strategy(self):
        """Order the candidates for search strategy."""
        for hnode, gnodes in self.candidates.iteritems():
            gnodes = sorted(gnodes, key=lambda node: -self.G.get_degree(node))
            self.candidates[hnode] = gnodes
        return

    def setup_search_strategy3(self):
        """Order the candidates for another search strategy."""
        gcov = dict()
        for hnode in self.H.get_nodes():
            gcov[hnode] = dict()
            for gnode in self.G.get_nodes():
                gcov[hnode][gnode] = 0

        for (hmap, gmap, _) in self.old_ones:
            for idx in xrange(len(hmap)):
                gcov[hmap[idx]][gmap[idx]] += 1

        print gcov

        for hnode, gnodes in self.candidates.iteritems():
            # first sort on degree, then on coverage
            gnodes = sorted(gnodes, key=lambda node: -self.G.get_degree(node))
            gnodes = sorted(gnodes, key=lambda gnode: gcov[hnode][gnode])
            self.candidates[hnode] = gnodes
            # print "   %s can be mapped to %s" % (hnode, gnodes)
        return

    def search(self, find_all):
        """Do the actual search."""
        hnodes = self.H.get_nodes()

        # first sort on degree, then on number of candidates
        hnodes = sorted(hnodes, key=lambda x: -self.H.get_degree(x))
        hnodes = sorted(hnodes, key=lambda x: len(self.candidates[x]))

        solutions = list()

        # This is our search tree
        T = Tree()
        self.searchtree = T

        M = Graph()

        # Now search!
        def search_step(T, hnodes, solutions, M, find_all):
            """Do one step in the search, more or less depth first."""
            DEBUG = True

            if DEBUG:
                print " -- Search one step --"

            hmap = list()
            gmap = list()

            parent = T
            value = parent.get_value()
            while value and value != "root":
                (hnode, gnode) = value
                hmap.append(hnode)
                gmap.append(gnode)
                parent = parent.get_parent()
                value = parent.get_value()

            if DEBUG and (self.ctr < 1000 or self.ctr % 250 == 0):
                print "   step %s" % self.ctr
                print "   %s items in hmap: %s" % (len(hmap), hmap)
                print "   %s items in gmap: %s" % (len(gmap), gmap)
            self.ctr += 1

            # TODO: hard coded limit -- use a counter instead?
            if self.ctr >= self.ctrmax:  # usually about a minute
                msg = "Tried for %s steps, giving up :-(" % self.ctr
                print msg
                raise GraphException(msg)

            if (datetime.now() - self.started).total_seconds() > self.maxtime:
                msg = "Mapping timeout! (after %s seconds)." % self.maxtime

                nowstring = datetime.now().strftime('%Y%m%d_%H%M%S')

                print
                print "WRITE TO FILE"
                print "   /tmp/%s-G.dot" % nowstring
                print "   /tmp/%s-H.dot" % nowstring
                print

                handle = open('/tmp/%s-G.dot' % nowstring, 'w')
                handle.write(self.G.to_dot())
                handle.flush()
                handle.close()

                handle = open('/tmp/%s-H.dot' % nowstring, 'w')
                handle.write(self.H.to_dot())
                handle.flush()
                handle.close()

                print msg
                raise GraphException(msg)

            # Remove all parent, grand parents and so on from the list
            # of unmapped nodes, if there are no unmapped nodes left
            # then we are done!

            for hnode in hmap:
                if hnode in hnodes:
                    hnodes.remove(hnode)

            if not hnodes:
                solutions.append((hmap, gmap, deepcopy(M)))
                if len(solutions) == 1:
                    print " --> first solution after %s steps!" % self.ctr
                return

            hnode = hnodes.pop(0)
            candidates = deepcopy(self.candidates[hnode])

            if DEBUG:
                print "   initial candidates of %s are %s." % (hnode,
                                                               candidates)

            M.add_node(hnode)
            for i in xrange(len(hmap)):
                if hnode in self.H.get_neighbours(hmap[i]):
                    for _ in xrange(self.H.get_num_edges_between(hmap[i],
                                                                 hnode)):
                        M.add_edge(hmap[i], hnode)

            for gnode in gmap:
                if gnode in candidates:
                    candidates.remove(gnode)

            if DEBUG:
                print "   hnode =",
                print hnode
                print "   nodes left=",
                print hnodes
                print "   candidates:",
                print candidates

            if not candidates:
                if DEBUG:
                    print "      kill this pair, due to no candidates"
                T.kill()

            for candidate in candidates:
                child = (hnode, candidate)
                if DEBUG:
                    print "   Added child (%s, %s) to search tree" % child
                T.add_child(child)

                for i in xrange(len(hmap)):
                    if hnode in self.H.get_neighbours(hmap[i]):
                        # if neighbour in H then it must be neighbour in G
                        if candidate not in self.G.get_neighbours(gmap[i]):
                            T[child].kill()
                        elif self.H.get_num_edges_between(hnode,
                                                          hmap[i]) > \
                            self.G.get_num_edges_between(candidate,
                                                         gmap[i]):
                            T[child].kill()

                if T[child].is_dead():
                    continue

                search_step(T[child], deepcopy(hnodes), solutions,
                            deepcopy(M), find_all)

                if not find_all and len(solutions) > 0:
                    return

            return

        search_step(T, hnodes, solutions, M, find_all)

        return solutions

    def to_dot(self, G, hmap, gmap, mapgraph):
        """Produce reasonable graphics."""
        output = "graph G{\n"

        print G.get_nodes()

        for node in G.get_nodes():
            clr = ''
            label = "undefined"
            if node not in gmap:
                clr = ', color="gray"'
                label = str(node)
            else:
                label = '%s\\n(%s)' % (hmap[gmap.index(node)], str(node))
            output += '    "%s" [label="%s"%s];\n' % (str(node), label, clr)

        # TODO: FIX BUG THAT COLORS ALL EDGES
        for (left, right) in G.get_edges():
            clr = ' [color="gray"]'

            if (left in gmap) and (right in gmap):
                hleft = hmap[gmap.index(left)]
                hright = hmap[gmap.index(right)]

                if (min(hleft, hright), max(hleft, hright)) in \
                   mapgraph.get_edges():
                    clr = ''

            output += "    %s -- %s%s;\n" % (left, right, clr)

        output += "}\n"
        return output

    def to_coverage_dot(self, G, gcov):
        """Produce a coverage graph based on coverage in gcov dict."""
        # four groups of styles
        #   0: unused nodes
        #   low: low 25%
        #   medium: middle 50%
        #   high: high 25%
        # also a style for edges between nodes with low usage,
        # as well as a style for normal usage

        if True:
            # OPTION 1 - gray scale
            node_style_0 = 'color="gray", style="dotted", fontcolor="gray"'
            node_style_low = 'color="gray"'
            node_style_medium = ''
            node_style_high = 'penwidth="2"'

            edge_style_0 = 'color="gray", style="dotted"'
            edge_style_low = 'color="gray"'
            edge_style_normal = ''
            edge_style_high = 'penwidth="2"'

        if False:
            # OPTION 2 - colors
            node_style_0 = 'color="salmon", style="dotted", fontcolor="salmon"'
            node_style_low = 'color="lightslateblue", style="dotted"'
            node_style_medium = 'color="slateblue"'
            node_style_high = 'color="darkslateblue", fontname="bold"'

            edge_style_low = 'color="lightsalmon"'
            edge_style_normal = ''

        coverages = [cov for cov in gcov.values() if cov > 0]

        cov1 = min(coverages)
        cov2 = max(coverages)

        if cov1 == cov2:
            node_style_low = node_style_medium
            node_style_high = node_style_medium
            # there might still me uncovered nodes

        styles = dict()
        for gnode in G.get_nodes():
            if gcov[gnode] == 0:
                styles[gnode] = node_style_0
            elif gcov[gnode] > cov1 + 0.75 * (cov2 - cov1):
                styles[gnode] = node_style_high
            elif gcov[gnode] > cov1 + 0.25 * (cov2 - cov1):
                styles[gnode] = node_style_medium
            else:
                styles[gnode] = node_style_low

        output = "graph G{\n"

        for gnode in G.get_nodes():
            output += '    "%s" [label="%s\\n(%s)"%s]\n' % (gnode, gnode,
                                                            gcov[gnode],
                                                            styles[gnode])
        for (left, right) in G.get_edges():
            clr = "[%s]" % edge_style_normal

            if gcov[left] == gcov[right] == 0:
                clr = "[%s]" % edge_style_0
            elif min([gcov[left], gcov[right]]) < cov1 + 0.25 * (cov2 - cov1):
                clr = "[%s]" % edge_style_low
            elif min([gcov[left], gcov[right]]) > cov1 + 0.75 * (cov2 - cov1):
                clr = "[%s]" % edge_style_high
            else:
                clr = ""

            output += "    %s -- %s %s;\n" % (left, right, clr)

        output += "}\n"
        return output

if __name__ == "__main__":
    pass
