"""
Simple graph class for playing around with subgraph isomorphism.

Author: Per Erik Strandberg, 2017

The source code in this bundle explores the subgraph isomorphism problem
Copyright (C) 2018 P. E. Strandberg, Westermo Research and Development AB

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""

from copy import deepcopy
from numpy import matrix, zeros


class GraphException(Exception):
    """Class of exceptions for Graph-related stuff."""

    pass


class Graph:
    """My graph class.

    Create a graph and, some nodes, get the nodes and count them:
        >>> G = Graph()
        >>> G.add_node(1)
        >>> G.add_node(2)
        >>> G.add_node(3)
        >>> G.add_node(4)
        >>> G.get_nodes() == [1, 2, 3, 4]
        True
        >>> G.get_num_nodes()
        4
        >>> G.get_num_edges_between(1, 2)
        0
        >>> G.get_degree(1)
        0

    In a graph without links the nodes have no neighbours
        >>> G.get_neighbours(2) == list()
        True

    A node cannot be added twice:
        >>> G.add_node(1) # doctest:+ELLIPSIS
        Traceback (most recent call last):
            ...
        GraphException: ...

    Nodes not in graph cannot be deleted
        >>> G.del_node(5) # doctest:+ELLIPSIS
        Traceback (most recent call last):
            ...
        GraphException: ...

    Let's add some edges, see how many we got and retrieve the edges:
        >>> G.add_edge(1, 2)
        >>> G.add_edge(2, 3)
        >>> G.add_edge(3, 1)
        >>> G.add_edge(3, 4)
        >>> G.add_edge(1, 4)
        >>> G.get_num_edges() == len(G.get_edges())
        True
        >>> G.get_num_edges_between(1, 2)
        1
        >>> G.get_degree(1)
        3

    The neighbours of 1 are now 2, 3 and 4:
        >>> G.get_neighbours(1) == [2, 3, 4]
        True

        >>> G.get_neighbours(4)
        [1, 3]

    Cannot add edges if node does not exist:
        >>> G.add_edge(5, 6) # doctest:+ELLIPSIS
        Traceback (most recent call last):
            ...
        GraphException: ...

    We might want a quick overview of this graph:
        >>> print G
        Graph with 4 nodes and 5 edges.

    Get string with dot-representation, for plotting with graphviz:
        >>> print G.to_dot()
        graph G{
            "1";
            "2";
            "3";
            "4";
            "1" -- "2";
            "1" -- "3";
            "1" -- "4";
            "2" -- "3";
            "3" -- "4";
        }
        <BLANKLINE>

    The degree sequence of a graph is really important later when
    considering unmappability:
        >>> print G.get_deg_seq()
        [3, 3, 2, 2]

    We can also not remove edges that does not exist:
        >>> G.del_edge(101, 102) # doctest:+ELLIPSIS
        Traceback (most recent call last):
            ...
        GraphException: ...

    Sometimes we need the adjacency matrix.
    Explanation: The second row has a 1 in the first and third column,
    this means that there is an edge between the second node and the
    first, and also an edge between the second node and the third.
        >>> A = G.get_matrix()
        >>> print A
        [[0 1 1 1]
         [1 0 1 0]
         [1 1 0 1]
         [1 0 1 0]]

    This might sound funky, but A to the power of something will
    reveal how many walks there are in G from a node to another.
    Since G has two loops of length three the sum of
    all diagonal elements of A ^ 3 divided by 6 will be 2.
        >>> A3 = A ** 3
        >>> A3.trace()[0, 0] == 2 * 6
        True

    If we remove an edge then the degree sequence changes:
        >>> G.del_edge(4, 3)
        >>> print G.get_deg_seq()
        [3, 2, 2, 1]

    The same is true when we delete nodes.
        >>> G.del_node(4)
        >>> print G.to_dot()
        graph G{
            "1";
            "2";
            "3";
            "1" -- "2";
            "1" -- "3";
            "2" -- "3";
        }
        <BLANKLINE>
        >>> print G.get_deg_seq()
        [2, 2, 2]

    When a node is removed we cannot get the neighbours:
        >>> G.get_neighbours(4) # doctest:+ELLIPSIS
        Traceback (most recent call last):
            ...
        GraphException: ...

    Let's add a tail to the graph:
        >>> G.add_node(5)
        >>> G.add_edge(5, 1)
        >>> for n in xrange(6, 10):
        ...     G.add_node(n)
        ...     G.add_edge(n-1, n)

    We now have a leaf, or end-node at then end of the tail:
        >>> print G.get_deg_seq()
        [3, 2, 2, 2, 2, 2, 2, 1]

    The neighbours of 7 are now 6 and 8:
        >>> G.get_neighbours(7) == [6, 8]
        True

    Let's add an extra edge between 7 and 8, and check how many we have:
        >>> G.get_num_edges_between(7, 8)
        1
        >>> G.add_edge(7, 8)
        >>> G.get_num_edges_between(7, 8)
        2
        >>> G.get_degree(7)
        3

    Let's delete a node in the middle to create an isolated node:
        >>> G.del_node(8)
        >>> print G.get_deg_seq()
        [3, 2, 2, 2, 2, 1, 0]

    The only neighbour of 7 is now 6:
        >>> G.get_neighbours(7) == [6, ]
        True

    There are now no edges left between 7 and 8:
        >>> G.get_num_edges_between(7, 8)
        0
        >>> G.get_degree(7)
        1
        >>> G.get_degree(9)
        0

    We can easily write the graph to file -- for processing with graphviz:
        >>> handle = open("mygraph.dot", "w")
        >>> handle.write(G.to_dot())
        >>> handle.close()

    """

    def __init__(self):
        """Just a constructor."""
        self._nodes = list()
        self._edges = list()
        self._degrees = dict()
        self._neighbours = dict()

        return

    def add_node(self, node):
        """Add a node to the graph."""
        if node not in self._nodes:
            self._nodes.append(node)
            self._nodes.sort()
            self._degrees[node] = 0
            self._neighbours[node] = list()
        else:
            raise GraphException("Cannot add node %s twice." % node)
        return

    def get_num_nodes(self):
        """Get number of nodes."""
        return len(self._nodes)

    def get_num_edges(self):
        """Get number of edges."""
        return len(self._edges)

    def get_num_edges_between(self, left, right):
        """Get number of edges between left and right."""
        (left, right) = (min(left, right), max(left, right))
        return self._edges.count((left, right))

    def get_nodes(self):
        """Get a (copy) of the nodes."""
        return deepcopy(self._nodes)

    def get_edges(self):
        """Get a (copy) of the edges."""
        return deepcopy(self._edges)

    def get_neighbours(self, node):
        """Get a (copy) of the neighbours."""
        if node not in self._nodes:
            raise GraphException("Graph has no node '%s'." % node)

        return deepcopy(self._neighbours[node])

    def get_degree(self, node):
        """Get the degree of a node."""
        return self._degrees[node]

    def add_edge(self, left, right):
        """Add an edge, if it turns out ok."""
        if (left in self._nodes) and (right in self._nodes):
            self._edges.append((min(left, right), max(left, right)))
            self._edges.sort()
            self._degrees[left] += 1
            self._degrees[right] += 1

            if right not in self._neighbours[left]:
                self._neighbours[left].append(right)
                self._neighbours[left].sort()
            if left not in self._neighbours[right]:
                self._neighbours[right].append(left)
                self._neighbours[right].sort()
        else:
            raise GraphException(
                "Cannot add edge (%s, %s) without nodes in graph." % (left,
                                                                      right))
        return

    def del_edge(self, left, right):
        """Delete an edge, if it turns out ok."""
        (left, right) = (min(left, right), max(left, right))
        if (left, right) not in self._edges:
            raise GraphException("Cannot remove edge not in graph.")

        self._edges.remove((left, right))
        self._degrees[left] -= 1
        self._degrees[right] -= 1

        if not (left, right) in self._edges:
            # there are no edges like the one we removed left in the graph,
            # so we need to update the neighbours
            self._neighbours[left].remove(right)
            self._neighbours[right].remove(left)

        return

    def del_node(self, node):
        """Delete a node, if it turns out ok."""
        if node not in self._nodes:
            raise GraphException("Cannot remove node not in graph.")

        # No need to care about degrees, it is taken care of in del_edge
        for (left, right) in deepcopy(self.get_edges()):
            if left == node:
                self.del_edge(left, right)
            elif right == node:
                self.del_edge(left, right)

        self._nodes.remove(node)
        del self._degrees[node]
        self._neighbours.pop(node)

        return

    def to_dot(self):
        """Print in format readable for graphviz."""
        output = "graph G{\n"

        for node in self._nodes:
            output += '    "%s";\n' % node

        for (left, right) in self._edges:
            output += '    "%s" -- "%s";\n' % (left, right)

        output += "}\n"
        return output

    def get_deg_seq(self):
        """Get the degree sequence."""
        return sorted(self._degrees.values(), reverse=True)

    def __str__(self):
        """String representation of graph."""
        return "Graph with %s nodes and %s edges." % (len(self._nodes),
                                                      len(self._edges))

    def get_matrix(self):
        """
        Get adjacency matrix as a NumPy matrix.

        A row in this vector can be though of as a list of node
        indices such that the node corresponding to this line is connected
        (adjancent) to. Columns work the same way, and it is
        symmetric: A[i, j] == A[j, i].

        See test case (above) and Wikipedia:
        https://en.wikipedia.org/wiki/Adjacency_matrix
        """
        A = matrix(zeros((len(self._nodes), len(self._nodes)), dtype=int))

        for (left, right) in self._edges:
            idx = self._nodes.index(left)
            jdx = self._nodes.index(right)

            A[idx, jdx] += 1
            A[jdx, idx] += 1

        return A


if __name__ == "__main__":
    import doctest
    res = doctest.testmod()
    print "Tested %s cases, %s failed." % (res.attempted, res.failed)
