"""
A class only useful for generating a bunch of test data.

Author: Per Erik Strandberg, 2017

The source code in this bundle explores the subgraph isomorphism problem
Copyright (C) 2018 P. E. Strandberg, Westermo Research and Development AB

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""

from network import Graph
from random import choice, seed, uniform
from string import ascii_uppercase as abcd


class TestGraphs:
    """Just generate some test graphs."""

    def make_empty_graphs(self, n_graphs):
        """Create a bunch of empty graphs."""
        for n in xrange(n_graphs):
            G = Graph()
            for m in xrange(n):
                G.add_node(m+1)
            self.graphs['N%d' % (n)] = G

        return

    def make_path_graphs(self, sizes):
        """Create a bunch of path graphs."""
        offset = 100
        for size in sizes:
            G = Graph()
            for n in xrange(1, size+1):
                G.add_node(n + offset)
                if n > 1:
                    G.add_edge(n-1 + offset, n + offset)
            self.graphs['P%s' % size] = G
            offset += 100
        return

    def make_cycle_graphs(self, sizes):
        """Create a bunch of cyclic graphs."""
        offset = 100
        for size in sizes:
            G = Graph()
            for n in xrange(1, size+1):
                G.add_node(n + offset)
                if n > 1:
                    G.add_edge(n-1 + offset, n + offset)
            G.add_edge(offset + 1, offset + size)
            self.graphs['C%s' % size] = G
            offset += 100
        return

    def make_complete_graphs(self, n_graphs):
        """Create a bunch of complete graphs."""
        # skip K0, and K1 -- same as N0, N1
        for n in xrange(2, n_graphs):
            G = Graph()
            for m in xrange(n):
                G.add_node(m+1)
                for i in xrange(m, 0, -1):
                    G.add_edge(i, m+1)
            self.graphs['K%d' % (n)] = G
        return

    def make_random_graphs(self, n_nodes, n_edges, myseed=9000):
        """Create a bunch of random graphs."""
        offset = 100

        idx = min(len(n_nodes), len(n_edges))

        while idx > 0:
            idx -= 1
            nodes = n_nodes[idx]
            edges = n_edges[idx]

            name = "R-%s-%s" % (nodes, edges)

            seed(myseed)
            G = Graph()
            for node in xrange(1, nodes+1):
                node += 1000 + offset
                G.add_node(node)

            for _ in xrange(edges):
                left = choice(G.get_nodes())
                right = choice(G.get_nodes())
                while left == right:
                    right = choice(G.get_nodes())

                G.add_edge(left, right)
            self.graphs[name] = G
            offset += 100
        return

    def make_t_graph(self, top, stem):
        """Create a bunch of T-shaped graphs."""
        # lets hope that top + stem is less than 26
        name = "T-%s-%s" % (top, stem)

        G = Graph()
        for idx in xrange(top):
            G.add_node(abcd[idx])

        for idx in xrange(1, top):
            G.add_edge(abcd[idx-1], abcd[idx])

        for idx in xrange(top, top+stem):
            G.add_node(abcd[idx])

        for idx in xrange(top+1, top+stem):
            G.add_edge(abcd[idx-1], abcd[idx])

        G.add_edge(abcd[top/2], abcd[top])

        self.graphs[name] = G
        return

    def make_grid_graphs(self, rows, cols, duplicity=1):
        """Make a bunch of grid graphs."""
        for i in rows:
            for j in cols:
                name = "G-%d-%d-%d" % (i, j, duplicity)
                G = Graph()

                # first row
                row = 0

                # add top left corner
                G.add_node(10000)
                for col in xrange(1, j):
                    node = 10000 + row*100 + col  # 11215 for row 12 col 15
                    G.add_node(node)

                    for _ in xrange(duplicity):
                        G.add_edge(node - 1, node)

                # remaining rows:
                for row in xrange(1, i):
                    col = 0
                    node = 10000 + row*100 + col
                    G.add_node(node)
                    for _ in xrange(duplicity):
                        G.add_edge(node - 100, node)
                    for col in xrange(1, j):
                        node = 10000 + row*100 + col  # 11215 for row 12 col 15
                        G.add_node(node)
                        for _ in xrange(duplicity):
                            G.add_edge(node - 1, node)
                            G.add_edge(node - 100, node)

                self.graphs[name] = G

    def make_scale_free_graphs(self, nodes, myseed):
        """Make scale free graphs by using a rich-get-richer metaphor.

        See chapter 18.3 in D. Easley & J. Kleinberg. (2010).
        Networks, crowds, and markets:
        Reasoning about a highly connected world.
        Cambridge University Press
        """
        p = 0.33

        # skip if nodes < 4
        # S1 = N1
        # S2 = P2
        # S3 = P3

        for n in nodes:
            # seed the random number generator,
            # so that we always build the same graphs
            seed(myseed)

            if n < 4:
                continue

            # take the first three here
            G = Graph()
            G.add_node(1)
            G.add_node(2)
            G.add_edge(1, 2)
            G.add_node(3)
            G.add_edge(3, choice([1, 2]))

            # now for the remaining nodes:
            for m in xrange(4, n+1):

                if uniform(0, 1) < p:
                    # With probability p, page j chooses a page i
                    # uniformly at random from among all earlier pages,
                    # and creates a link to this page i.
                    target = choice(G.get_nodes())
                    G.add_node(m)
                    G.add_edge(target, m)
                else:
                    # With probability (1-p), page j instead chooses a
                    # page i uniformly at random from among all
                    # earlier pages, and creates a link to the page
                    # that i points to.
                    tmp = choice(G.get_nodes())
                    target = choice(G.get_neighbours(tmp))
                    G.add_node(m)
                    G.add_edge(target, m)

            self.graphs["S%d-%d" % (n, myseed)] = G

        return

    def make_graphs_from_report(self):
        """Make graphs drawn as examples in paper."""
        #     1 - 2      G0
        #     |   |
        # 3 - 4 - 5 - 6 - 7 - 8
        #     |   |
        #     9 - X
        G0 = Graph()

        for node in xrange(1, 11):
            G0.add_node(node)
        G0.add_edge(1, 2)
        G0.add_edge(1, 4)
        G0.add_edge(2, 5)
        G0.add_edge(3, 4)
        G0.add_edge(4, 5)
        G0.add_edge(4, 9)
        G0.add_edge(5, 6)
        G0.add_edge(5, 10)
        G0.add_edge(6, 7)
        G0.add_edge(7, 8)
        G0.add_edge(9, 10)
        self.graphs["G0"] = G0

        #     1       H1
        #     |
        # 2 - 3 - 4 - 5
        #     |
        #     6
        H1 = Graph()
        for node in xrange(201, 207):
            H1.add_node(node)
        H1.add_edge(201, 203)
        H1.add_edge(202, 203)
        H1.add_edge(203, 204)
        H1.add_edge(204, 205)
        H1.add_edge(203, 206)
        self.graphs["H1"] = H1

        #     1       H2
        #     |
        # 2 - 3 - 4 - 5 - 6
        #     |
        #     7
        H2 = Graph()
        for node in xrange(301, 308):
            H2.add_node(node)
        H2.add_edge(301, 303)
        H2.add_edge(302, 303)
        H2.add_edge(303, 304)
        H2.add_edge(304, 305)
        H2.add_edge(305, 306)
        H2.add_edge(303, 307)
        self.graphs["H2"] = H2

        # H3 is the same as a 2, 3 grid graph -- let's make it any way.
        # 1 - 2      H3
        # |   |
        # 3 - 4
        H3 = Graph()
        for node in xrange(401, 405):
            H3.add_node(node)
        H3.add_edge(401, 402)
        H3.add_edge(401, 403)
        H3.add_edge(402, 404)
        H3.add_edge(403, 404)
        self.graphs["H3"] = H3

        return

    def __init__(self):
        """Constructor that generates the test graphs."""
        self.graphs = dict()

        self.make_empty_graphs(10)
        self.make_complete_graphs(10)
        self.make_grid_graphs([2, 3, 6, 9, 20, 30], [4, 8, 11, 15, 19, 30, 40])
        self.make_grid_graphs([2, 3, 6, 9], [4, 8, 11, 15, 19], 2)
        self.make_scale_free_graphs([4, 5, 10, 20, 25, 30, 35, 40, 100], 42)
        self.make_scale_free_graphs([4, 5, 10, 20, 25, 30, 35, 40, 100], 9000)
        self.make_graphs_from_report()
        self.make_path_graphs([1, 2, 3, 4, 5, 6, 7, 8, 10, 20, 30])
        self.make_cycle_graphs([4, 5, 6, 7, 10, 20, 30])
        self.make_random_graphs([5, 10, 20, 40], [10, 30, 50, 100])
        self.make_t_graph(3, 2)

        return

    def print_graphs(self):
        """Print all graphs."""
        for (name, G) in self.graphs.iteritems():
            print name,
            print G.get_num_nodes(),
            print G.get_num_edges()
            print G.to_dot()
            print
        return


if __name__ == "__main__":
    TG = TestGraphs()

    for myname in sorted(TG.graphs.keys()):
        print "Graph",
        print myname,
        print "has",
        print TG.graphs[myname].get_num_nodes(),
        print "nodes, and",
        print TG.graphs[myname].get_num_edges(),
        print "edges."
