"""
Just a simple tree structure.

Author: Per Erik Strandberg, 2017

The source code in this bundle explores the subgraph isomorphism problem
Copyright (C) 2018 P. E. Strandberg, Westermo Research and Development AB

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""

from copy import deepcopy


class Tree:
    """
    Class for a tree. Has data, a parent and possibly a bunch of kids.

    For simplicity we will think of the children as also being trees.

    Let's create a tree, and add some kids to the root:
        >>> T = Tree()
        >>> T.add_child(1)
        >>> T.add_child(2)
        >>> T.add_child(3)

    We can ask a tree for the value it has, sometimes we need this:
        >>> T.get_value() == "root"
        True
        >>> T[2].get_value()
        2

    The tree now has one level (root is "below ground"):
        >>> T.get_levels()
        1

    OK, now we want to add 2, 3 and 4 to the first child.
    These are not the same 2 and 3 as the children of the root.
    Also, let's add some children to the second child of the root.
        >>> T[1].add_child(2)
        >>> T[1].add_child(3)
        >>> T[1].add_child(4)
        >>> T[2].add_child(1)
        >>> T[2].add_child(3)

    Now, let's take a look at one of these nodes.
        >>> my_tree_node = T[1]
        >>> my_tree_node.get_parent() == T
        True
        >>> my_tree_node.get_children() == [2, 3, 4]
        True
        >>> my_tree_node[3].get_parent() == my_tree_node
        True

    The tree now has two levels:
        >>> T.get_levels()
        2

    The tree can display itself
        >>> print T
        Tree with 9 nodes, in 2 levels.

    Let's get a third level and see if we get the right information.
        >>> T[1][3].add_child(5)

    We should have 10 nodes by now:
        >>> len(T.get_nodes())
        10

    The tree now has two levels:
        >>> T.get_levels()
        3

    The tree can display itself
        >>> print T
        Tree with 10 nodes, in 3 levels.

    And let's display the tree in dot format.
        >>> print T.to_dot()
        graph T{
            "root" [label="root", shape="diamond"];
            "1" [label="1"];
            "2" [label="2"];
            "3" [label="3"];
            "1-2" [label="2"];
            "1-3" [label="3"];
            "1-4" [label="4"];
            "2-1" [label="1"];
            "2-3" [label="3"];
            "1-3-5" [label="5"];
            "root" -- "1";
            "root" -- "2";
            "root" -- "3";
            "1" -- "1-2";
            "1" -- "1-3";
            "1" -- "1-4";
            "2" -- "2-1";
            "2" -- "2-3";
            "1-3" -- "1-3-5";
        }
        <BLANKLINE>

    We can also see if a branch is dead or not:
        >>> T[1][3][5].is_dead()
        False
        >>> T[1][3][5].kill()
        >>> T[1][3][5].is_dead()
        True

    """

    def __init__(self, value=None, parent=None):
        """Constructor assumes that this is the root."""
        if parent is None:
            self._value = "root"
        else:
            self._value = value

        self._parent = parent
        self._dead = False
        self._children = dict()
        return

    def is_dead(self):
        """Return true if branch is dead."""
        return self._dead

    def kill(self):
        """Kill this branch."""
        self._dead = True

    def get_parent(self):
        """Get the parent."""
        return self._parent

    def get_value(self):
        """Get the parent."""
        return deepcopy(self._value)

    def get_children(self):
        """Get the children of this node."""
        # TODO: Is the sorting bad?
        return deepcopy(sorted(self._children.keys()))

    def add_child(self, value):
        """Add a child tree to this tree with self as parent."""
        if self.is_dead():
            raise Exception("Cannot add nodes to a dead tree!!!")

        t = Tree(value, self)
        self._children[value] = t
        return

    def __getitem__(self, key):
        """Accessor for convenience."""
        return self._children[key]

    def get_nodes(self):
        """Get the nodes of this tree."""
        nodes = list()
        nodes.append(self)

        for child in sorted(self._children.values()):
            for node in child.get_nodes():
                nodes.append(node)

        return nodes

    def get_levels(self):
        """Get the number of levels in the tree."""
        # empty tree, only root node
        if len(self._children.values()) == 0:
            return 0

        return 1 + max([c.get_levels() for c in self._children.values()])

    def __str__(self):
        """String representation of a Tree."""
        return "Tree with %d nodes, in %d levels." % (len(self.get_nodes()),
                                                      self.get_levels())

    def to_dot(self):
        """Create a nice string representation of the Tree, for graphviz."""
        output = "graph T{\n"

        this_level = list()
        this_level.append(self)

        edges = list()

        while this_level:
            next_level = list()

            for node in this_level:

                label = "---undefined---"

                if node.get_value() == 'root':
                    label = "root"
                else:
                    label = node.get_value()
                    parent = node.get_parent()

                    while parent.get_value() != "root":
                        label = "%s-%s" % (parent.get_value(), label)
                        parent = parent.get_parent()

                clr = ""
                if node.get_value() == "root":
                    clr = ', shape="diamond"'
                if node.is_dead():
                    clr = ', color="gray"'
                output += '    "%s" [label="%s"%s];\n' % (label,
                                                          node.get_value(),
                                                          clr)

                # store next level
                for child in node._children.values():
                    if label == "root":
                        edges.append(("root", child.get_value()))
                    else:
                        edges.append((label,
                                      "%s-%s" % (label, child.get_value())))
                    next_level.append(child)

            this_level = next_level

        for (left, right) in edges:
            output += '    "%s" -- "%s";\n' % (left, right)

        output += "}\n"
        return output


if __name__ == "__main__":
    import doctest
    res = doctest.testmod()
    print "Tested %s cases, %s failed." % (res.attempted, res.failed)
